import datetime
import numpy
import segyio
import segyio._segyio as _segyio


def default_text_header(iline, xline, offset):
    lines = {
        1: "DATE %s" % datetime.date.today().isoformat(),
        2: "AN INCREASE IN AMPLITUDE EQUALS AN INCREASE IN ACOUSTIC IMPEDANCE",
        3: "Written by libsegyio (python)",
        11: "TRACE HEADER POSITION:",
        12: "  INLINE BYTES %03d-%03d    | OFFSET BYTES %03d-%03d" % (iline, iline + 4, int(offset), int(offset) + 4),
        13: "  CROSSLINE BYTES %03d-%03d |" % (xline, xline + 4),
        15: "END EBCDIC HEADER",
    }
    rows = segyio.create_text_header(lines)
    rows = bytearray(rows, 'ascii')  # mutable array of bytes
    rows[-1] = 128  # \x80 -- Unsure if this is really required...
    return bytes(rows)  # immutable array of bytes that is compatible with strings


def create(filename, spec):
    """Create a new segy file.

    Create a new segy file with the geometry and properties given by `spec`.
    This enables creating SEGY files from your data. The created file supports
    all segyio modes, but has an emphasis on writing. The spec must be
    complete, otherwise an exception will be raised. A default, empty spec can
    be created with `segyio.spec()`.

    Very little data is written to the file, so just calling create is not
    sufficient to re-read the file with segyio. Rather, every trace header and
    trace must be written to for the file to be considered complete.

    Create should be used together with python's `with` statement. This ensure
    the data is written. Please refer to the examples.

    Args:
        filename (str): Path to file to open.
        spec (:obj: `spec`): Structure of the segy file.

    Examples:
        Create a file::
            >>> spec = segyio.spec()
            >>> spec.ilines  = [1, 2, 3, 4]
            >>> spec.xlines  = [11, 12, 13]
            >>> spec.samples = list(range(50))
            >>> spec.sorting = 2
            >>> spec.format  = 1
            >>> with segyio.create(path, spec) as f:
            ...     ## fill the file with data
            ...

        Copy a file, but shorten all traces by 50 samples::
            >>> with segyio.open(srcpath) as src:
            ...     spec = segyio.spec()
            ...     spec.sorting = src.sorting
            ...     spec.format = src.format
            ...     spec.samples = src.samples[:len(src.samples) - 50]
            ...     spec.ilines = src.ilines
            ...     spec.xline = src.xlines
            ...     with segyio.create(dstpath, spec) as dst:
            ...         dst.text[0] = src.text[0]
            ...         dst.bin = src.bin
            ...         dst.header = src.header
            ...         dst.trace = src.trace
    :rtype: segyio.SegyFile
    """
    f = segyio.SegyFile(filename, "w+")

    f._samples       = numpy.asarray(spec.samples, dtype = numpy.single)
    f._ext_headers   = spec.ext_headers
    f._bsz           = _segyio.trace_bsize(len(f.samples))

    txt_hdr_sz       = _segyio.textheader_size()
    bin_hdr_sz       = _segyio.binheader_size()
    f._tr0           = txt_hdr_sz + bin_hdr_sz + (spec.ext_headers * txt_hdr_sz)
    f._sorting       = spec.sorting
    f._fmt           = int(spec.format)
    f._offsets       = numpy.copy(numpy.asarray(spec.offsets, dtype = numpy.intc))
    f._tracecount    = len(spec.ilines) * len(spec.xlines) * len(spec.offsets)

    f._il            = int(spec.iline)
    f._ilines        = numpy.copy(numpy.asarray(spec.ilines, dtype=numpy.intc))

    f._xl            = int(spec.xline)
    f._xlines        = numpy.copy(numpy.asarray(spec.xlines, dtype=numpy.intc))

    line_metrics = _segyio.init_line_metrics(f.sorting, f.tracecount, len(f.ilines), len(f.xlines), len(f.offsets))

    f._iline_length = line_metrics['iline_length']
    f._iline_stride = line_metrics['iline_stride']

    f._xline_length = line_metrics['xline_length']
    f._xline_stride = line_metrics['xline_stride']

    f.text[0] = default_text_header(f._il, f._xl, segyio.TraceField.offset)
    f.bin     = {
        3213: f.tracecount,
        3217: 4000,
        3221: len(f.samples),
        3225: f.format,
        3505: f.ext_headers,
    }

    return f



import numpy

import segyio


def open(filename, mode="r", iline=189, xline=193, strict = True):
    """Open a segy file.

    Opens a segy file and tries to figure out its sorting, inline numbers,
    crossline numbers, and offsets, and enables reading and writing to this
    file in a simple manner.

    For reading, the access mode "r" is preferred. All write operations will
    raise an exception. For writing, the mode "r+" is preferred (as "rw" would
    truncate the file). The modes used are standard C file modes; please refer
    to that documentation for a complete reference.

    Open should be used together with python's `with` statement. Please refer
    to the examples. When the `with` statement is used the file will
    automatically be closed when the routine completes or an exception is
    raised.

    By default, segyio tries to open in 'strict' mode. This means the file will
    be assumed to represent a geometry with consistent inline, crosslines and
    offsets. If strict is False, segyio will still try to establish a geometry,
    but it won't abort if it fails. When in non-strict mode is opened,
    geometry-dependent modes such as iline will raise an error.

    Args:
        filename (str): Path to file to open.
        mode (str, optional): File access mode, defaults to "r".
        iline (TraceField): Inline number field in the trace headers. Defaults
                            to 189 as per the SEGY specification.
        xline (TraceField): Crossline number field in the trace headers.
                            Defaults to 193 as per the SEGY specification.
        strict (bool, optional): Abort if a geometry cannot be inferred.
                                 Defaults to True.

    Examples:
        Open a file in read-only mode::
            >>> with segyio.open(path, "r") as f:
            ...     print(f.ilines)
            ...

        Open a file in read-write mode::
            >>> with segyio.open(path, "r+") as f:
            ...     f.trace = np.arange(100)
            ...

        Open two files at once::
            >>> with segyio.open(path) as src, segyio.open(path, "r+") as dst:
            ...     dst.trace = src.trace # copy all traces from src to dst
            ...
    :rtype: segyio.SegyFile
    """
    f = segyio.SegyFile(filename, mode, iline, xline)

    try:
        metrics = segyio._segyio.init_metrics(f.xfd, f.bin.buf)

        f._tr0 = metrics['trace0']
        f._fmt = metrics['format']
        f._bsz = metrics['trace_bsize']
        f._tracecount = metrics['trace_count']
        f._ext_headers = (f._tr0 - 3600) // 3200  # should probably be from C

        dt = segyio.tools.dt(f, fallback_dt = 4000.0)
        t0 = f.header[0][segyio.TraceField.DelayRecordingTime]
        sample_count = metrics['sample_count']
        f._samples = numpy.array([t0 + i * dt for i in range(sample_count)],
                                 dtype = numpy.single) / 1000.0

    except:
        f.close()
        raise
    try:
        cube_metrics = segyio._segyio.init_cube_metrics(f.xfd,
                                                        iline,
                                                        xline,
                                                        f.tracecount,
                                                        f._tr0,
                                                        f._bsz)
        f._sorting   = cube_metrics['sorting']
        iline_count  = cube_metrics['iline_count']
        xline_count  = cube_metrics['xline_count']
        offset_count = cube_metrics['offset_count']
        metrics.update(cube_metrics)

        line_metrics = segyio._segyio.init_line_metrics(f.sorting,
                                                        f.tracecount,
                                                        iline_count,
                                                        xline_count,
                                                        offset_count)

        f._iline_length = line_metrics['iline_length']
        f._iline_stride = line_metrics['iline_stride']

        f._xline_length = line_metrics['xline_length']
        f._xline_stride = line_metrics['xline_stride']

        f._ilines  = numpy.zeros(iline_count,  dtype = numpy.intc)
        f._xlines  = numpy.zeros(xline_count,  dtype = numpy.intc)
        f._offsets = numpy.zeros(offset_count, dtype = numpy.intc)
        segyio._segyio.init_indices(f.xfd, metrics, f.ilines, f.xlines, f.offsets)

        if numpy.unique(f.ilines).size != f.ilines.size:
            raise ValueError( "Inlines inconsistent - expect all inlines to be unique")

        if numpy.unique(f.xlines).size != f.xlines.size:
            raise ValueError( "Crosslines inconsistent - expect all crosslines to be unique")

    except:
        if not strict:
            f._ilines  = None
            f._xlines  = None
            f._offsets = None
        else:
            f.close()
            raise

    return f



import numpy as np
import segyio
from segyio._raw_trace import RawTrace


class Trace:

    def __init__(self, file):
        self._file = file
        """:type: segyio.file"""

    def __getitem__(self, index, buf=None):
        if isinstance(index, tuple):
            return self.__getitem__(index[0], index[1])

        buf = self._trace_buffer(buf)

        if isinstance(index, int):
            if not 0 <= abs(index) < len(self):
                raise IndexError("Trace %d not in range (-%d,%d)", (index, len(self), len(self)))

            return self._readtr(index, buf)

        elif isinstance(index, slice):
            def gen():
                for i in range(*index.indices(len(self))):
                    yield self._readtr(i, buf)

            return gen()

        else:
            raise TypeError("Key must be int, slice, (int,np.ndarray) or (slice,np.ndarray)")

    def __setitem__(self, index, val):
        if not 0 <= abs(index) < len(self):
            raise IndexError("Trace %d not in range (-%d,%d)", (index, len(self), len(self)))

        if not isinstance(val, np.ndarray):
            raise TypeError("Value must be numpy.ndarray")

        if val.dtype != np.single:
            raise TypeError("Numpy array must be of type single")

        shape = (len(self._file.samples),)

        if val.shape[0] < shape[0]:
            raise TypeError("Array wrong shape. Expected minimum %s, was %s" % (shape, val.shape))

        if isinstance(index, int):
            self._writetr(index, val)

        elif isinstance(index, slice):
            for i, buf in range(*index.indices(len(self))), val:
                self._writetr(i, val)

        else:
            raise KeyError("Wrong shape of index")

    def __len__(self):
        return self._file.tracecount

    def __iter__(self):
        return self[:]

    def __repr__(self):
        return "Trace(traces = {}, samples = {})".format(len(self), self._file.samples)

    def _trace_buffer(self, buf=None):
        shape = self._file.samples.shape

        if buf is None:
            buf = np.empty(shape=shape, dtype=np.single)
        elif not isinstance(buf, np.ndarray):
            raise TypeError("Buffer must be None or numpy.ndarray")
        elif buf.dtype != np.single:
            buf = np.empty(shape=shape, dtype=np.single)

        return buf

    def _readtr(self, traceno, buf=None):
        buf = self._trace_buffer(buf)

        tracecount = self._file.tracecount
        trace0 = self._file._tr0
        bsz = self._file._bsz
        fmt = self._file._fmt
        smp = len(self._file.samples)
        return segyio._segyio.read_trace(self._file.xfd, traceno, tracecount, buf, trace0, bsz, fmt, smp)

    def _writetr(self, traceno, buf):
        self.write_trace(traceno, buf, self._file)

    @classmethod
    def write_trace(cls, traceno, buf, segy):
        """
        :type traceno: int
        :type buf: ?
        :type segy: segyio.SegyFile
        """
        segyio._segyio.write_trace(segy.xfd, traceno,
                                   buf,
                                   segy._tr0, segy._bsz,
                                   segy._fmt, len(segy.samples))

    @property
    def raw(self):
        """ :rtype: segyio.RawTrace """
        return RawTrace(self)



import segyio
from segyio import BinField
from segyio import TraceField


class Field:
    def __init__(self, buf, write, field_type, traceno=None):
        self.buf = buf
        self.traceno = traceno
        self._get_field = segyio._segyio.get_field
        self._set_field = segyio._segyio.set_field
        self._field_type = field_type
        self._write = write

    def __getitem__(self, field):

        # add some structure so we can always iterate over fields
        if isinstance(field, int) or isinstance(field, self._field_type):
            field = [field]

        d = {self._field_type(f): self._get_field(self.buf, f) for f in field}

        # unpack the dictionary. if header[field] is requested, a
        # plain, unstructed output is expected, but header[f1,f2,f3]
        # yields a dict
        if len(d) == 1:
            return d.popitem()[1]

        return d

    def __setitem__(self, field, val):
        self._set_field(self.buf, field, val)
        self._write(self.buf, self.traceno)

    def update(self, value):
        buf = self.buf
        if isinstance(value, dict):
            for k, v in value.items():
                self._set_field(buf, int(k), v)
        else:
            buf = value.buf

        self._write(buf, self.traceno)

    @classmethod
    def binary(cls, segy):
        try:
            buf = segyio._segyio.read_binaryheader(segy.xfd)
        except IOError:
            # the file was probably newly created and the binary header hasn't
            # been written yet.  if this is the case we want to try and write
            # it. if the file was broken, permissions were wrong etc writing
            # will fail too
            buf = segyio._segyio.empty_binaryheader()

        def wr(buf, *_):
            segyio._segyio.write_binaryheader(segy.xfd, buf)

        return Field(buf, write=wr, field_type=BinField)

    @classmethod
    def trace(cls, buf, traceno, segy):
        if traceno >= segy.tracecount:
            raise IndexError("Trace number out of range: 0 <= %d < %d" % (traceno, segy.tracecount))

        if buf is None:
            buf = segyio._segyio.empty_traceheader()

        try:
            segyio._segyio.read_traceheader(segy.xfd, traceno, buf, segy._tr0, segy._bsz)
        except IOError:
            # the file was probably newly created and the trace header hasn't
            # been written yet.  if this is the case we want to try and write
            # it. if the file was broken, permissions were wrong etc writing
            # will fail too
            pass

        def wr(buf, traceno):
            segyio._segyio.write_traceheader(segy.xfd, traceno, buf, segy._tr0, segy._bsz)

        return Field(buf, traceno=traceno, write=wr, field_type=TraceField)

    def __repr__(self):
        def assigned(x):
            return x != BinField.Unassigned1 and x != BinField.Unassigned2

        fields = filter(assigned, self._field_type.enums())
        return self[fields].__repr__()



import itertools

import segyio._segyio as _segyio

try:
    from itertools import izip as zip
except ImportError:  # will be 3.x series
    pass


class Line:
    """ Line mode for traces and trace headers. Internal.

    The _line class provides an interface for line-oriented operations. The
    line reading operations themselves are not streaming - it's assumed that
    when a line is queried it's somewhat limited in size and will comfortably
    fit in memory, and that the full line is interesting. This also applies to
    line headers; however, all returned values support the iterable protocol so
    they work fine together with the streaming bits of this library.

    _line should not be instantiated directly by users, but rather returned
    from the iline/xline properties of file or from the header mode. Any
    direct construction of this should be considered an error.
    """

    def __init__(self, segy, length, stride, lines, other_lines, buffn, readfn, writefn, name):
        self.segy = segy
        self.len = length
        self.stride = stride
        self.lines = lines
        self.other_lines = other_lines
        self.name = name
        self.buffn = buffn
        self.readfn = readfn
        self.writefn = writefn

    def __repr__(self):
        return "Line(direction = {}, length = {})".format(self.name, self.len)

    def _index(self, lineno, offset):
        """ :rtype: int"""
        offs = self.segy.offsets

        if offset is None:
            offset = 0
        else:
            try:
                offset = next(i for i, x in enumerate(offs) if x == offset)
            except StopIteration:
                try:
                    int(offset)
                except TypeError:
                    raise TypeError("Offset must be int or slice")

                raise KeyError("Unknkown offset {}".format(offset))

        try:
            lineno = int(lineno)
        except TypeError:
            raise TypeError("Must be int or slice")

        trace0 = _segyio.fread_trace0(lineno, len(self.other_lines), self.stride, len(offs), self.lines, self.name)
        return offset + trace0

    def _indices(self, lineno, offset):
        """ :rtype: tuple[collections.Iterable, collections.Iterable]"""
        offsets = self.segy.offsets
        if offset is None:
            offset = offsets[0]

        if not isinstance(lineno, slice):
            lineno = slice(lineno, lineno + 1, 1)

        if not isinstance(offset, slice):
            offset = slice(offset, offset + 1, 1)

        lineno = self._sanitize_slice(lineno, self.lines)
        offset = self._sanitize_slice(offset, offsets)

        offs, lns = set(self.segy.offsets), set(self.lines)

        orng = range(*offset.indices(offsets[-1] + 1))
        lrng = range(*lineno.indices(self.lines[-1] + 1))

        return filter(lns.__contains__, lrng), filter(offs.__contains__, orng)

    # in order to support [:end] syntax, we must make sure
    # start has a non-None value. lineno.indices() would set it
    # to 0, but we don't know if that's a reasonable value or
    # not. If start is None we set it to the first line
    def _sanitize_slice(self, s, source):
        if all((s.start, s.stop, s.step)):
            return s

        start, stop, step = s.start, s.stop, s.step
        increasing = step is None or step > 0

        if start is None:
            start = source[0] if increasing else source[-1]

        if stop is None:
            stop = source[-1] + 1 if increasing else source[0] - 1

        return slice(start, stop, step)

    def _get(self, lineno, offset, buf):
        """ :rtype: numpy.ndarray"""
        t0 = self._index(lineno, offset)
        return self.readfn(t0, self.len, self.stride, buf)

    def _get_iter(self, lineno, off, buf):
        """ :rtype: collections.Iterable[numpy.ndarray]"""

        for line, offset in itertools.product(*self._indices(lineno, off)):
            yield self._get(line, offset, buf)

    def __getitem__(self, lineno, offset=None):
        """ :rtype: numpy.ndarray|collections.Iterable[numpy.ndarray]"""
        buf = self.buffn()

        if isinstance(lineno, tuple):
            lineno, offset = lineno

        if isinstance(lineno, slice) or isinstance(offset, slice):
            return self._get_iter(lineno, offset, buf)

        return self._get(lineno, offset, buf)

    def __setitem__(self, lineno, val):
        offset = None
        if isinstance(lineno, tuple):
            lineno, offset = lineno

        if isinstance(lineno, slice) or isinstance(offset, slice):
            lines, offsets = self._indices(lineno, offset)

            indices = itertools.product(*self._indices(lineno, offset))
            for (line, offset), x in zip(indices, val):
                t0 = self._index(line, offset)
                self.writefn(t0, self.len, self.stride, x)

        else:
            t0 = self._index(lineno, offset)
            self.writefn(t0, self.len, self.stride, val)

    def __len__(self):
        return len(self.lines)

    def __iter__(self):
        buf = self.buffn()
        for i in self.lines:
            yield self._get(i, None, buf)



from segyio import Enum


class SegySampleFormat(Enum):
    IBM_FLOAT_4_BYTE = 1
    SIGNED_INTEGER_4_BYTE = 2
    SIGNED_SHORT_2_BYTE = 3
    FIXED_POINT_WITH_GAIN_4_BYTE = 4
    IEEE_FLOAT_4_BYTE = 5
    NOT_IN_USE_1 = 6
    NOT_IN_USE_2 = 7
    SIGNED_CHAR_1_BYTE = 8



import numpy as np
import segyio

class RawTrace(object):
    def __init__(self, trace):
        self.trace = trace

    def __getitem__(self, index):
        """ :rtype: numpy.ndarray """
        buf = None
        if isinstance(index, slice):
            f = self.trace._file
            start, stop, step = index.indices(f.tracecount)
            mstart, mstop = min(start, stop), max(start, stop)
            length = max(0, (mstop - mstart + (step - (1 if step > 0 else -1))))
            buf = np.zeros(shape = (length, len(f.samples)), dtype = np.single)

        return self.trace._readtr(index, buf)

    def __repr__(self):
        return self.trace.__repr__() + ".raw"



from segyio import Enum


class TraceField(Enum):
    TRACE_SEQUENCE_LINE = 1
    TRACE_SEQUENCE_FILE = 5
    FieldRecord = 9
    TraceNumber = 13
    EnergySourcePoint = 17
    CDP = 21
    CDP_TRACE = 25
    TraceIdentificationCode = 29
    NSummedTraces = 31
    NStackedTraces = 33
    DataUse = 35
    offset = 37
    ReceiverGroupElevation = 41
    SourceSurfaceElevation = 45
    SourceDepth = 49
    ReceiverDatumElevation = 53
    SourceDatumElevation = 57
    SourceWaterDepth = 61
    GroupWaterDepth = 65
    ElevationScalar = 69
    SourceGroupScalar = 71
    SourceX = 73
    SourceY = 77
    GroupX = 81
    GroupY = 85
    CoordinateUnits = 89
    WeatheringVelocity = 91
    SubWeatheringVelocity = 93
    SourceUpholeTime = 95
    GroupUpholeTime = 97
    SourceStaticCorrection = 99
    GroupStaticCorrection = 101
    TotalStaticApplied = 103
    LagTimeA = 105
    LagTimeB = 107
    DelayRecordingTime = 109
    MuteTimeStart = 111
    MuteTimeEND = 113
    TRACE_SAMPLE_COUNT = 115
    TRACE_SAMPLE_INTERVAL = 117
    GainType = 119
    InstrumentGainConstant = 121
    InstrumentInitialGain = 123
    Correlated = 125
    SweepFrequencyStart = 127
    SweepFrequencyEnd = 129
    SweepLength = 131
    SweepType = 133
    SweepTraceTaperLengthStart = 135
    SweepTraceTaperLengthEnd = 137
    TaperType = 139
    AliasFilterFrequency = 141
    AliasFilterSlope = 143
    NotchFilterFrequency = 145
    NotchFilterSlope = 147
    LowCutFrequency = 149
    HighCutFrequency = 151
    LowCutSlope = 153
    HighCutSlope = 155
    YearDataRecorded = 157
    DayOfYear = 159
    HourOfDay = 161
    MinuteOfHour = 163
    SecondOfMinute = 165
    TimeBaseCode = 167
    TraceWeightingFactor = 169
    GeophoneGroupNumberRoll1 = 171
    GeophoneGroupNumberFirstTraceOrigField = 173
    GeophoneGroupNumberLastTraceOrigField = 175
    GapSize = 177
    OverTravel = 179
    CDP_X = 181
    CDP_Y = 185
    INLINE_3D = 189
    CROSSLINE_3D = 193
    ShotPoint = 197
    ShotPointScalar = 201
    TraceValueMeasurementUnit = 203
    TransductionConstantMantissa = 205
    TransductionConstantPower = 209
    TransductionUnit = 211
    TraceIdentifier = 213
    ScalarTraceHeader = 215
    SourceType = 217
    SourceEnergyDirectionMantissa = 219
    SourceEnergyDirectionExponent = 223
    SourceMeasurementMantissa = 225
    SourceMeasurementExponent = 229
    SourceMeasurementUnit = 231
    UnassignedInt1 = 233
    UnassignedInt2 = 237


"""
segy modes

Welcome to segyio.segy. Here you will find references and examples for the
various segy modes and how to interact with segy files. To start interacting
with files, please refer to the segyio.open and segyio.create documentation, by
typing `help(segyio.open)` or `help(segyio.create)`.

The primary way of obtaining a file instance is calling segyio.open. When you
have a file instance you can interact with it as described in this module.

The explanations and examples here are meant as a quick guide and reference.
You can also have a look at the example programs that are distributed with
segyio which you can find in the examples directory or where your distribution
installs example programs.
"""
import itertools
import numpy as np

from segyio._header import Header
from segyio._line import Line
from segyio._trace import Trace
from segyio._field import Field
import segyio._segyio as _segyio

from segyio.tracesortingformat import TraceSortingFormat

try:
    from itertools import izip as zip
except ImportError:  # will be 3.x series
    pass

try: xrange
except NameError: pass
else: range = xrange

class SegyFile(object):

    _unstructured_errmsg = "File opened in unstructured mode."

    def __init__(self, filename, mode, iline=189, xline=193):
        """
        Constructor, internal.
        """

        self._filename = filename
        self._mode = mode
        self._il = iline
        self._xl = xline

        # property value holders
        self._ilines = None
        self._xlines = None
        self._tracecount = None
        self._sorting = None
        self._offsets = None
        self._ext_headers = None
        self._samples = None

        # private values
        self._iline_length = None
        self._iline_stride = None
        self._xline_length = None
        self._xline_stride = None
        self._fmt = None
        self._tr0 = None
        self._bsz = None

        self.xfd = _segyio.open(filename, mode)

        super(SegyFile, self).__init__()

    def __str__(self):
        f = "SegyFile {}:".format(self._filename)

        if self.unstructured:
            il =  "  inlines: None"
            xl =  "  crosslines: None"
            of =  "  offsets: None"
        else:
            il =  "  inlines: {} [{}, {}]".format(len(self.ilines), self.ilines[0], self.ilines[-1])
            xl =  "  crosslines: {} [{}, {}]".format(len(self.xlines), self.xlines[0], self.xlines[-1])
            of =  "  offsets: {} [{}, {}]".format(len(self.offsets), self.offsets[0], self.offsets[-1])

        tr =  "  traces: {}".format(self.tracecount)
        sm =  "  samples: {}".format(self.samples)
        fmt = "  float representation: {}".format(self.format)

        props = [f, il, xl, tr, sm]

        if self.offsets is not None and len(self.offsets) > 1:
            props.append(of)

        props.append(fmt)
        return '\n'.join(props)


    def __repr__(self):
        return "SegyFile('{}', '{}', iline = {}, xline = {})".format(
                        self._filename, self._mode, self._il, self._xl)

    def __enter__(self):
        """Internal.
        :rtype: segyio.segy.SegyFile
        """
        return self

    def __exit__(self, type, value, traceback):
        """Internal."""
        self.close()

    def flush(self):
        """Flush a file - write the library buffers to disk.

        This method is mostly useful for testing.

        It is not necessary to call this method unless you want to observe your
        changes while the file is still open. The file will automatically be
        flushed for you if you use the `with` statement when your routine is
        completed.

        Examples:
            Flush::
                >>> with segyio.open(path) as f:
                ...     # write something to f
                ...     f.flush()
        """
        _segyio.flush(self.xfd)

    def close(self):
        """Close the file.

        This method is mostly useful for testing.

        It is not necessary to call this method if you're using the `with`
        statement, which will close the file for you.
        """
        _segyio.close(self.xfd)

    def mmap(self):
        """Memory map the file
        :rtype: bool

        Memory map the file. This is an advanced feature for speed and
        optimization; however, it is no silver bullet. If your file is smaller
        than the memory available on your system this will likely result in
        faster reads and writes, especially for line modes. However, if the
        file is very large, or memory is very pressured, this optimization
        might cause overall system slowdowns. However, if you're opening the
        same file from many different instances of segyio then memory mapping
        may significantly reduce the memory pressure.

        If this call returns true, the file is memory mapped. If memory mapping
        was build-time disabled or is not available for your platform this call
        always return false. If the memory mapping is unsuccessful you can keep
        using segyio - reading and writing falls back on non-memory mapped
        features.

        Examples:
            Memory map::
            >>> mapped = f.mmap()
            >>> if mapped: print( "File is memory mapped!" )
            >>> # keep using segyio as per usual
            >>> print( f.trace[10] )
        """
        return _segyio.mmap(self.xfd)

    @property
    def sorting(self):
        """ :rtype: int """
        return self._sorting

    @property
    def tracecount(self):
        """ :rtype: int """
        return self._tracecount

    @property
    def samples(self):
        """ :rtype: int """
        return self._samples

    @property
    def offsets(self):
        """ :rtype: numpy.ndarray"""
        return self._offsets

    @property
    def ext_headers(self):
        """ :rtype: int """
        return self._ext_headers

    @property
    def unstructured(self):
        return self.ilines is None

    @property
    def header(self):
        """ Interact with segy in header mode.

        This mode gives access to reading and writing functionality of headers,
        both in individual (trace) mode and line mode. Individual headers are
        accessed via generators and are not read from or written to disk until
        the generator is realised and the header in question is used. Supports
        python slicing (which yields a generator), as well as direct lookup.

        Examples:
            Reading a field in a trace::
                >>> import segyio
                >>> f = segyio.open("filename")
                >>> f.header[10][TraceField.offset]

            Writing a field in a trace::
                >>> f.header[10][TraceField.offset] = 5

            Copy a header from another header::
                >>> f.header[28] = f.header[29]

            Reading multiple fields in a trace. If raw, numerical offsets are
            used they must align with the defined byte offsets by the SEGY
            specification::
                >>> f.header[10][TraceField.offset, TraceField.INLINE_3D]
                >>> f.header[10][37, 189]

            Write multiple fields in a trace::
                >>> f.header[10] = { 37: 5, TraceField.INLINE_3D: 2484 }

            Iterate over headers and gather line numbers::
                >>> [h[TraceField.INLINE_3D] for h in f.header]
                >>> [h[25, 189] for h in f.header]

            Write field in all headers::
                >>> for h in f.header:
                ...     h[37] = 1
                ...     h = { TraceField.offset: 1, 2484: 10 }
                ...

            Read a field in 10 first headers::
                >>> [h[25] for h in f.header[0:10]]

            Read a field in every other header::
                >>> [h[37] for h in f.header[::2]]

            Write a field in every other header::
                >>> for h in f.header[::2]:
                ...     h = { TraceField.offset : 2 }
                ...

            Cache a header:
                >>> h = f.header[12]
                >>> x = foo()
                >>> h[37] = x

            A convenient way for operating on all headers of a file is to use the
            default full-file range.  It will write headers 0, 1, ..., n, but uses
            the iteration specified by the right-hand side (i.e. can skip headers
            etc).

            If the right-hand-side headers are exhausted before all the destination
            file headers the writing will stop, i.e. not all all headers in the
            destination file will be written to.

            Copy headers from file g to file f:
                >>> f = segyio.open("path to file")
                >>> g = segyio.open("path to another file")
                >>> f.header = g.header

            Set offset field::
                >>> f.header = { TraceField.offset: 5 }

            Copy every 12th header from the file g to f's 0, 1, 2...::
                >>> f.header = g.header[::12]
                >>> f.header[0] == g.header[0]
                True
                >>> f.header[1] == g.header[12]
                True
                >>> f.header[2] == g.header[2]
                False
                >>> f.header[2] == g.header[24]
                True

            The header mode can also be accessed with line addressing, which
            supports all of iline and xline's indexing features.

            Rename the iline 3 to 4::
                >>> f.header.iline[3][TraceField.INLINE_3D] = 4
                >>> # please note that rewriting the header won't update the
                >>> # file's interpretation of the file until you reload it, so
                >>> # the new iline 4 will be considered iline 3 until the file
                >>> # is reloaded

            Set offset line 3 offset 3 to 5::
                >>> f.header.iline[3, 3] = { TraceField.offset: 5 }

        :rtype: segyio._header.Header
        """
        return Header(self)

    @header.setter
    def header(self, val):
        if isinstance(val, Field) or isinstance(val, dict):
            val = itertools.repeat(val)

        h, buf = self.header, None
        for i, v in zip(range(self.tracecount), val):
            h[i, buf] = v

    def attributes(self, field):
        """ File-wide attribute (header word) reading

        A range-oriented function that reads some attribute for all the
        specified headers file-wide. Supports index lookup, slices and
        numpy-style list-of-indices.

        Examples:
            Read all unique sweep frequency end::
                >>> end = segyio.TraceField.SweepFrequencyEnd
                >>> sfe = np.unique(f.attributes( end )[:])

            Discover the first traces of each unique sweep frequency end::
                >>> end = segyio.TraceField.SweepFrequencyEnd
                >>> attrs = f.attributes(end)
                >>> sfe, tracenos = np.unique(attrs[:], return_index = True)

            Scatter plot group x/y-coordinates with SFEs (using matplotlib)::
                >>> end = segyio.TraceField.SweepFrequencyEnd
                >>> attrs = f.attributes(end)
                >>> _, tracenos = np.unique(attrs[:], return_index = True)
                >>> gx = f.attributes(segyio.TraceField.GroupX)[tracenos]
                >>> gy = f.attributes(segyio.TraceField.GroupY)[tracenos]
                >>> scatter(gx, gy)
        """
        class attr:
            def __getitem__(inner, rng):
                try: iter(rng)
                except TypeError: pass
                else: return inner._getitem_list(rng)

                if not isinstance(rng, slice):
                    rng = slice(rng, rng + 1, 1)

                traces = self.tracecount
                start, stop, step = rng.indices(traces)
                attrs = np.empty(len(range(*rng.indices(traces))), dtype = np.intc)
                return _segyio.field_forall(self.xfd, attrs,
                                            start, stop, step, field,
                                            self._tr0, self._bsz)

            def _getitem_list(inner, xs):
                if not isinstance(xs, np.ndarray):
                    xs = np.asarray(xs, dtype = np.intc)

                xs = xs.astype(dtype = np.intc, order = 'C', copy = False)
                attrs = np.empty(len(xs), dtype = np.intc)
                return _segyio.field_foreach(self.xfd, attrs, xs, field,
                                          self._tr0, self._bsz)

        return attr()


    @property
    def trace(self):
        """ Interact with segy in trace mode.

        This mode gives access to reading and writing functionality for traces.
        The primary data type is the numpy ndarray. Traces can be accessed
        individually or with python slices, and writing is done via assignment.

        All examples use `np` for `numpy`.

        Examples:
            Read all traces in file f and store in a list::
                >>> l = [np.copy(tr) for tr in f.trace]

            Do numpy operations on a trace::
                >>> tr = f.trace[10]
                >>> tr = np.transpose(tr)
                >>> tr = tr * 2
                >>> tr = tr - 100
                >>> avg = np.average(tr)

            Do numpy operations on every other trace::
                >>> for tr in f.trace[::2]:
                ...     print( np.average(tr) )
                ...

            Traverse traces in reverse::
                >>> for tr in f.trace[::-1]:
                ...     print( np.average(tr) )
                ...

            Double every trace value and write to disk. Since accessing a trace
            gives a numpy value, to write to the respective trace we need its index::
                >>> for i, tr in enumerate(f.trace):
                ...     tr = tr * 2
                ...     f.trace[i] = tr
                ...

            Reuse an array for memory efficiency when working with indices.
            When using slices or full ranges this is done for you::
                >>> tr = None
                >>> for i in range(100):
                ...     tr = f.trace[i, tr]
                ...     tr = tr * 2
                ...     print(np.average(tr))
                ...

            Read a value directly from a file. The second [] is numpy access
            and supports all numpy operations, including negative indexing and
            slicing::
                >>> f.trace[0][0]
                1490.2
                >>> f.trace[0][1]
                1490.8
                >>> f.trace[0][-1]
                1871.3
                >>> f.trace[-1][100]
                1562.0

            Trace mode supports len(), returning the number of traces in a
            file::
                >>> len(f.trace)
                300

            Convenient way for setting traces from 0, 1, ... n, based on the
            iterable set of traces on the right-hand-side.

            If the right-hand-side traces are exhausted before all the destination
            file traces the writing will stop, i.e. not all all traces in the
            destination file will be written.

            Copy traces from file f to file g::
                >>> f.trace = g.trace.

            Copy first half of the traces from g to f::
                >>> f.trace = g.trace[:len(g.trace)/2]

            Fill the file with one trace (filled with zeros)::
                >>> tr = np.zeros(f.samples)
                >>> f.trace = itertools.repeat(tr)

            For advanced users: sometimes you want to load the entire segy file
            to memory and apply your own structural manipulations or operations
            on it. Some segy files are very large and may not fit, in which
            case this feature will break down. This is an optimisation feature;
            using it should generally be driven by measurements.

            Read the first 10 traces::
                >>> f.trace.raw[0:10]

            Read *all* traces to memory::
                >>> f.trace.raw[:]

            Read every other trace to memory::
                >>> f.trace.raw[::2]
        """

        return Trace(self)

    @trace.setter
    def trace(self, val):
        tr = self.trace
        for i, v in zip(range(len(tr)), val):
            tr[i] = v

    def _shape_buffer(self, shape, buf):
        if buf is None:
            return np.empty(shape=shape, dtype=np.single)
        if not isinstance(buf, np.ndarray):
            return buf
        if buf.dtype != np.single:
            return np.empty(shape=shape, dtype=np.single)
        if buf.shape[0] == shape[0]:
            return buf
        if buf.shape != shape and buf.size == np.prod(shape):
            return buf.reshape(shape)
        return buf

    def _line_buffer(self, length, buf=None):
        shape = (length, len(self.samples))
        return self._shape_buffer(shape, buf)

    def _fread_line(self, trace0, length, stride, buf):
        offsets = len(self.offsets)
        return _segyio.read_line(self.xfd, trace0,
                                length, stride, offsets,
                                buf,
                                self._tr0, self._bsz,
                                self._fmt, len(self.samples))

    @property
    def ilines(self):
        """ :rtype: numpy.ndarray"""
        return self._ilines

    @property
    def iline(self):
        """ Interact with segy in inline mode.

        This mode gives access to reading and writing functionality for inlines.
        The primary data type is the numpy ndarray. Inlines can be accessed
        individually or with python slices, and writing is done via assignment.
        Note that accessing inlines uses the line numbers, not their position,
        so if a files has inlines [2400..2500], accessing line [0..100] will be
        an error. Note that each line is returned as a numpy array, meaning
        accessing the intersections of the inline and crossline is 0-indexed.

        Additionally, the iline mode has a concept of offsets, which is useful
        when dealing with prestack files. Offsets are accessed via so-called
        sub indexing, meaning iline[10, 4] will give you line 10 at offset 4.
        Please note that offset, like lines, are accessed via their numbers,
        not their indices. If your file has the offsets [150, 250, 350, 450]
        and the lines [2400..2500], you can access the third offset with
        [2403,350]. Please refer to the examples for more details. If no offset
        is specified, segyio will give you the first.

        Examples:
            Read an inline::
                >>> il = f.iline[2400]

            Copy every inline into a list::
                >>> l = [np.copy(x) for x in f.iline]

            The number of inlines in a file::
                >>> len(f.iline)

            Numpy operations on every other inline::
                >>> for line in f.iline[::2]:
                ...     line = line * 2
                ...     avg = np.average(line)
                ...     print(avg)
                ...

            Read inlines up to 2430::
                >>> for line in f.iline[:2430]:
                ...     print(np.average(line))
                ...

            Copy a line from file g to f::
                >>> f.iline[2400] = g.iline[2834]

            Copy lines from the first line in g to f, starting at 2400,
            ending at 2410 in f::
                >>> f.iline[2400:2411] = g.iline

            Convenient way for setting inlines, from left-to-right as the inline
            numbers are specified in the file.ilines property, from an iterable
            set on the right-hand-side.

            If the right-hand-side inlines are exhausted before all the destination
            file inlines the writing will stop, i.e. not all all inlines in the
            destination file will be written.

            Copy inlines from file f to file g::
                >>> f.iline = g.iline.

            Copy first half of the inlines from g to f::
                >>> f.iline = g.iline[:g.ilines[len(g.ilines)/2]]

            Copy every other inline from a different file::
                >>> f.iline = g.iline[::2]

            Accessing offsets work the same way as accessing lines, and slicing
            is supported as well. When doing range-based offset access, the
            lines will be generated offsets-first, i.e equivalent to:
            [(line1 off1), (line1 off2), (line1 off3), (line2 off1), ...]
            or the double for loop::
                >>> for line in lines:
                ...     for off in offsets:
                ...         yield (line, off)
                ...

            Copy all lines at all offsets::
                >>> [np.copy(x) for x in f.iline[:,:]]

            Print all line 10's offsets::
                >>> print(f.iline[10,:])

            Numpy operations at every line at offset 120::
                >>> for line in f.iline[:, 120]:
                ...     line = line * 2
                ...     print(np.average(line))

            Copy every other line and offset::
                >>> map(np.copy, f.iline[::2, ::2])

            Print offsets in reverse::
                >>> for line in f.iline[:, ::-1]:
                ...     print(line)

            Copy all offsets [200, 250, 300, 350, ...] in the range [200, 800)
            for all ilines [2420,2460)::
                >>> [np.copy(x) for x in f.iline[2420:2460, 200:800:50]]

            Copy every third offset from f to g::
                >>> g.iline[:,:] = f.iline[:,::3]

            Copy an iline from f to g at g's offset 200::
                >>> g.iline[12, 200] = f.iline[21]
        """

        if self.unstructured:
            raise ValueError(self._unstructured_errmsg)

        il_len, il_stride = self._iline_length, self._iline_stride
        lines = self.ilines
        other_lines = self.xlines
        buffn = lambda x=None: self._line_buffer(il_len, x)
        readfn = self._fread_line

        def writefn(t0, length, step, val):
            val = buffn(val)
            step *= len(self.offsets)
            for i, v in zip(range(t0, t0 + (step * length), step), val):
                Trace.write_trace(i, v, self)

        return Line(self, il_len, il_stride, lines, other_lines, buffn, readfn, writefn, "Inline")

    @iline.setter
    def iline(self, value):
        self.iline[:] = value

    @property
    def xlines(self):
        """ :rtype: numpy.ndarray"""
        return self._xlines

    @property
    def xline(self):
        """ Interact with segy in crossline mode.

        This mode gives access to reading and writing functionality for crosslines.
        The primary data type is the numpy ndarray. crosslines can be accessed
        individually or with python slices, and writing is done via assignment.
        Note that accessing crosslines uses the line numbers, not their position,
        so if a files has crosslines [1400..1450], accessing line [0..100] will be
        an error. Note that each line is returned as a numpy array, meaning
        accessing the intersections of the inline and crossline is 0-indexed.

        Additionally, the xline mode has a concept of offsets, which is useful
        when dealing with prestack files. Offsets are accessed via so-called
        sub indexing, meaning xline[10, 4] will give you line 10 at offset 4.
        Please note that offset, like lines, are accessed via their numbers,
        not their indices. If your file has the offsets [100, 200, 300, 400]
        and the lines [1400..1450], you can access the second offset with
        [1421,300]. Please refer to the examples for more details. If no offset
        is specified, segyio will give you the first.

        Examples:
            Read an crossline::
                >>> il = f.xline[1400]

            Copy every crossline into a list::
                >>> l = [np.copy(x) for x in f.xline]

            The number of crosslines in a file::
                >>> len(f.xline)

            Numpy operations on every third crossline::
                >>> for line in f.xline[::3]:
                ...     line = line * 6
                ...     avg = np.average(line)
                ...     print(avg)
                ...

            Read crosslines up to 1430::
                >>> for line in f.xline[:1430]:
                ...     print(np.average(line))
                ...

            Copy a line from file g to f::
                >>> f.xline[1400] = g.xline[1603]

            Copy lines from the first line in g to f, starting at 1400,
            ending at 1415 in f::
                >>> f.xline[1400:1416] = g.xline


            Convenient way for setting crosslines, from left-to-right as the crossline
            numbers are specified in the file.xlines property, from an iterable
            set on the right-hand-side.

            If the right-hand-side crosslines are exhausted before all the destination
            file crosslines the writing will stop, i.e. not all all crosslines in the
            destination file will be written.

            Copy crosslines from file f to file g::
                >>> f.xline = g.xline.

            Copy first half of the crosslines from g to f::
                >>> f.xline = g.xline[:g.xlines[len(g.xlines)/2]]

            Copy every other crossline from a different file::
                >>> f.xline = g.xline[::2]

            Accessing offsets work the same way as accessing lines, and slicing
            is supported as well. When doing range-based offset access, the
            lines will be generated offsets-first, i.e equivalent to:
            [(line1 off1), (line1 off2), (line1 off3), (line2 off1), ...]
            or the double for loop::
                >>> for line in lines:
                ...     for off in offsets:
                ...         yield (line, off)
                ...

            Copy all lines at all offsets::
                >>> [np.copy(x) for x in f.xline[:,:]]

            Print all line 10's offsets::
                >>> print(f.xline[10,:])

            Numpy operations at every line at offset 120::
                >>> for line in f.xline[:, 120]:
                ...     line = line * 2
                ...     print(np.average(line))

            Copy every other line and offset::
                >>> map(np.copy, f.xline[::2, ::2])

            Print offsets in reverse::
                >>> for line in f.xline[:, ::-1]:
                ...     print(line)

            Copy all offsets [200, 250, 300, 350, ...] in the range [200, 800)
            for all xlines [2420,2460)::
                >>> [np.copy(x) for x in f.xline[2420:2460, 200:800:50]]

            Copy every third offset from f to g::
                >>> g.xline[:,:] = f.xline[:,::3]

            Copy an xline from f to g at g's offset 200::
                >>> g.xline[12, 200] = f.xline[21]
        """

        if self.unstructured:
            raise ValueError(self._unstructured_errmsg)

        xl_len, xl_stride = self._xline_length, self._xline_stride
        lines = self.xlines
        other_lines = self.ilines
        buffn = lambda x=None: self._line_buffer(xl_len, x)
        readfn = self._fread_line

        def writefn(t0, length, step, val):
            val = buffn(val)
            step *= len(self.offsets)
            for i, v in zip(range(t0, t0 + step * length, step), val):
                Trace.write_trace(i, v, self)

        return Line(self, xl_len, xl_stride, lines, other_lines, buffn, readfn, writefn, "Crossline")

    @xline.setter
    def xline(self, value):
        self.xline[:] = value

    def _depth_buffer(self, buf=None):
        il_len = self._iline_length
        xl_len = self._xline_length

        if self.sorting == TraceSortingFormat.CROSSLINE_SORTING:
            shape = (il_len, xl_len)
        elif self.sorting == TraceSortingFormat.INLINE_SORTING:
            shape = (xl_len, il_len)
        else:
            raise RuntimeError("Unexpected sorting type")

        return self._shape_buffer(shape, buf)

    @property
    def fast(self):
        """ Access the 'fast' dimension

        This mode yields iline or xline mode, depending on which one is laid
        out "faster", i.e. the line with linear disk layout. Use this mode if
        the inline/crossline distinction isn't as interesting as traversing in
        a fast manner (typically when you want to apply a function to the whole
        file, line-by-line).
        """
        if self.sorting == TraceSortingFormat.INLINE_SORTING:
            return self.iline
        elif self.sorting == TraceSortingFormat.CROSSLINE_SORTING:
            return self.xline
        else:
            raise RuntimeError("Unknown sorting.")

    @property
    def slow(self):
        """ Access the 'slow' dimension

        This mode yields iline or xline mode, depending on which one is laid
        out "slower", i.e. the line with strided disk layout. Use this mode if
        the inline/crossline distinction isn't as interesting as traversing in
        the slower direction.
        """
        if self.sorting == TraceSortingFormat.INLINE_SORTING:
            return self.xline
        elif self.sorting == TraceSortingFormat.CROSSLINE_SORTING:
            return self.iline
        else:
            raise RuntimeError("Unknown sorting.")

    @property
    def depth_slice(self):
        """ Interact with segy in depth slice mode.

        This mode gives access to reading and writing functionality for depth slices.
        The primary data type is the numpy ndarray. Depth slices can be accessed
        individually or with python slices, and writing is done via assignment.
        Note that each slice is returned as a numpy array, meaning
        accessing the values of the slice is 0-indexed.

        Examples:
            Read a depth slice:
                >>> il = f.depth_slice[199]

            Copy every depth slice into a list::
                >>> l = [np.copy(x) for x in f.depth_slice]

            The number of depth slices in a file::
                >>> len(f.depth_slice)

            Numpy operations on every third depth slice::
                >>> for depth_slice in f.depth_slice[::3]:
                ...     depth_slice = depth_slice * 6
                ...     avg = np.average(depth_slice)
                ...     print(avg)
                ...

            Read depth_slices up to 250::
                >>> for depth_slice in f.depth_slice[:250]:
                ...     print(np.average(depth_slice))
                ...

            Copy a line from file g to f::
                >>> f.depth_slice[4] = g.depth_slice[19]

            Copy lines from the first line in g to f, starting at 10,
            ending at 49 in f::
                >>> f.depth_slice[10:50] = g.depth_slice


            Convenient way for setting depth slices, from left-to-right as the depth slices
            numbers are specified in the file.depth_slice property, from an iterable
            set on the right-hand-side.

            If the right-hand-side depth slices are exhausted before all the destination
            file depth slices the writing will stop, i.e. not all all depth slices in the
            destination file will be written.

            Copy depth slices from file f to file g::
                >>> f.depth_slice = g.depth_slice.

            Copy first half of the depth slices from g to f::
                >>> f.depth_slice = g.depth_slice[:g.samples/2]]

            Copy every other depth slices from a different file::
                >>> f.depth_slice = g.depth_slice[::2]
        """

        if self.unstructured:
            raise ValueError(self._unstructured_errmsg)

        indices = np.asarray(list(range(len(self.samples))), dtype=np.intc)
        other_indices = np.asarray([0], dtype=np.intc)
        buffn = self._depth_buffer

        slice_trace_count = self._iline_length * self._xline_length
        offsets = len(self.offsets)
        tr0 = self._tr0
        bsz = self._bsz
        fmt = self._fmt

        def readfn(depth, length, stride, buf):
            _segyio.depth_slice(self.xfd, depth,
                                slice_trace_count, offsets,
                                buf,
                                tr0, bsz,
                                fmt,
                                len(self.samples))
            return buf

        def writefn(depth, length, stride, val):
            val = buffn(val)

            buf_view = val.reshape(self._iline_length * self._xline_length)

            for i, trace_buf in enumerate(self.trace):
                trace_buf[depth] = buf_view[i]
                self.trace[i] = trace_buf

        return Line(self, len(self.samples), 1, indices, other_indices, buffn, readfn, writefn, "Depth")

    @depth_slice.setter
    def depth_slice(self, value):
        self.depth_slice[:] = value

    @property
    def text(self):
        """ Interact with segy in text mode.

        This mode gives access to reading and writing functionality for textual
        headers.

        The primary data type is the python string. Reading textual headers is
        done via [], and writing is done via assignment. No additional
        structure is built around the textual header, so everything is treated
        as one long string without line breaks.

        Examples:
            Print the textual header::
                >>> print(f.text[0])

            Print the first extended textual header::
                >>> print(f.text[1])

            Write a new textual header::
                >>> f.text[0] = make_new_header()

            Print a textual header line-by-line::
                >>> # using zip, from the zip documentation
                >>> text = f.text[0]
                >>> lines = map(''.join, zip( *[iter(text)] * 80))
                >>> for line in lines:
                ...     print(line)
                ...
        """

        class TextHeader:

            def __getitem__(inner, index):
                if index > self.ext_headers:
                    raise IndexError("Textual header %d not in file" % index)

                return _segyio.read_textheader(self.xfd, index)

            def __setitem__(inner, index, val):
                if index > self.ext_headers:
                    raise IndexError("Textual header %d not in file" % index)

                _segyio.write_textheader(self.xfd, index, val)

            def __repr__(inner):
                return "Text(external_headers = {})".format(self.ext_headers)

            def __str__(inner):
                return '\n'.join(map(''.join, zip(*[iter(inner[0])] * 80)))

        return TextHeader()

    @property
    def bin(self):
        """ Interact with segy in binary mode.

        This mode gives access to reading and writing functionality for the
        binary header. Please note that using numeric binary offsets uses the
        offset numbers from the specification, i.e. the first field of the
        binary header starts at 3201, not 1. If you're using the enumerations
        this is handled for you.

        Examples:
            Copy a header from file g to file f::
                >>> f.bin = g.bin

            Reading a field in a trace::
                >>> traces_per_ensemble = f.bin[3213]

            Writing a field in a trace::
                >>> f.bin[BinField.Traces] = 5

            Reading multiple fields::
                >>> d = f.bin[BinField.Traces, 3233]

            Copy a field from file g to file f::
                >>> f.bin[BinField.Format] = g.bin[BinField.Format]

            Copy full binary from file f to file g::
                >>> f.bin = g.bin

            Copy multiple fields from file f to file g::
                >>> f.bin = g.bin[BinField.Traces, 3233]

            Write field in binary header via dict::
                >>> f.bin = { BinField.Traces: 350 }

            Write multiple fields in a trace::
                >>> f.bin = { 3213: 5, BinField.SweepFrequencyStart: 17 }
        """

        return Field.binary(self)

    @bin.setter
    def bin(self, value):
        self.bin.update(value)

    @property
    def format(self):
        d = {
            1: "4-byte IBM float",
            2: "4-byte signed integer",
            3: "2-byte signed integer",
            4: "4-byte fixed point with gain",
            5: "4-byte IEEE float",
            8: "1-byte signed char"
        }

        class fmt:
            def __int__(inner):
                return self._fmt

            def __str__(inner):
                if not self._fmt in d:
                    return "Unknown format"

                return d[self._fmt]

        return fmt()

class spec:
    def __init__(self):
        self.iline = 189
        self.ilines = None
        self.xline = 193
        self.xlines = None
        self.offsets = [1]
        self.samples = None
        self.tracecount = None
        self.ext_headers = 0
        self.format = None
        self.sorting = None
        self.t0 = 1111.0



import itertools

import segyio
from segyio._line import Line
from segyio._field import Field

try:
    from itertools import izip as zip
except ImportError:  # will be 3.x series
    pass


class Header(object):
    def __init__(self, segy):
        self.segy = segy

    @staticmethod
    def _header_buffer(buf=None):
        if buf is None:
            buf = segyio._segyio.empty_traceheader()
        return buf

    def __getitem__(self, traceno, buf=None):
        if isinstance(traceno, tuple):
            return self.__getitem__(traceno[0], traceno[1])

        if isinstance(traceno, slice):
            gen_buf = self._header_buffer(buf)

            def gen():
                for i in range(*traceno.indices(self.segy.tracecount)):
                    yield self.__getitem__(i, gen_buf)

            return gen()

        return Field.trace(buf, traceno=traceno, segy=self.segy)

    def __setitem__(self, traceno, val):
        buf = None

        # library-provided loops can re-use a buffer for the lookup, even in
        # __setitem__, so we might need to unpack the tuple to reuse the buffer
        if isinstance(traceno, tuple):
            buf = traceno[1]
            traceno = traceno[0]

        self.__getitem__(traceno, buf).update(val)

    def __iter__(self):
        return self[:]

    def __repr__(self):
        return "Header(traces = {})".format(self.segy.samples)

    def readfn(self, t0, length, stride, buf):
        def gen():
            start = t0
            step = stride * len(self.segy.offsets)
            stop = t0 + (length * step)
            for i in range(start, stop, step):
                yield Field.trace(buf, traceno=i, segy=self.segy)

        return gen()

    def writefn(self, t0, length, stride, val):
        start = t0
        stride *= len(self.segy.offsets)
        stop = t0 + (length * stride)

        if isinstance(val, Field) or isinstance(val, dict):
            val = itertools.repeat(val)

        for i, x in zip(range(start, stop, stride), val):
            self[i] = x

    @property
    def iline(self):
        """:rtype: Line"""
        segy = self.segy
        length = segy._iline_length
        stride = segy._iline_stride
        lines = segy.ilines
        other_lines = segy.xlines
        buffn = self._header_buffer

        return Line(segy, length, stride, lines, other_lines, buffn, self.readfn, self.writefn, "Inline")

    @iline.setter
    def iline(self, value):
        """Write iterables to lines

        Examples:
            Supports writing to *all* crosslines via assignment, regardless of
            data source and format. Will respect the sample size and structure
            of the file being assigned to, so if the argument traces are longer
            than that of the file being written to the surplus data will be
            ignored. Uses same rules for writing as `f.iline[i] = x`.
        """
        for i, src in zip(self.segy.ilines, value):
            self.iline[i] = src

    @property
    def xline(self):
        """:rtype: Line"""
        segy = self.segy
        length = segy._xline_length
        stride = segy._xline_stride
        lines = segy.xlines
        other_lines = segy.ilines
        buffn = self._header_buffer

        return Line(segy, length, stride, lines, other_lines, buffn, self.readfn, self.writefn, "Crossline")

    @xline.setter
    def xline(self, value):
        """Write iterables to lines

        Examples:
            Supports writing to *all* crosslines via assignment, regardless of
            data source and format. Will respect the sample size and structure
            of the file being assigned to, so if the argument traces are longer
            than that of the file being written to the surplus data will be
            ignored. Uses same rules for writing as `f.xline[i] = x`.
        """

        for i, src in zip(self.segy.xlines, value):
            self.xline[i] = src



#!/usr/bin/env python

from distutils.core import setup, Extension

_segyio = Extension('segyio._segyio',
                    sources=['python/segyio/_segyio.c', 'lib/src/segy.c'],
                    include_dirs=['lib/include', 'lib/src'],
                    define_macros=[('HAVE_MMAP', 1), ('HAVE_NETINET_IN_H', 1), ('HAVE_SYS_STAT_H', 1)],
                    extra_compile_args=['-std=c99'])

long_description = """
=======
SEGY IO
=======

Introduction
------------

Segyio is a small LGPL licensed C library for easy interaction with SEG Y
formatted seismic data, with language bindings for Python and Matlab. Segyio is
an attempt to create an easy-to-use, embeddable, community-oriented library for
seismic applications. Features are added as they are needed; suggestions and
contributions of all kinds are very welcome.

Feature summary
---------------
 * A low-level C interface with few assumptions; easy to bind to other
   languages.
 * Read and write binary and textual headers.
 * Read and write traces, trace headers.
 * Easy to use and native-feeling python interface with numpy integration.

Project goals
-------------

Segyio does necessarily attempt to be the end-all of SEG-Y interactions;
rather, we aim to lower the barrier to interacting with SEG-Y files for
embedding, new applications or free-standing programs.

Additionally, the aim is not to support the full standard or all exotic (but
correctly) formatted files out there. Some assumptions are made, such as:

 * All traces in a file are assumed to be of the same sample size.
 * It is assumed all lines have the same number of traces.

The writing functionality in Segyio is largely meant to *modify* or adapt
files. A file created from scratch is not necessarily a to-spec SEG-Y file, as
we only necessarily write the header fields segyio needs to make sense of the
geometry. It is still highly recommended that SEG-Y files are maintained and
written according to specification, but segyio does not mandate this.

"""

setup(name='SegyIO',
      version='1.0.4.1',
      description='IO library for SEG-Y files',
      long_description=long_description,
      author='Statoil ASA',
      author_email='ert@statoil.com',
      url='https://github.com/Statoil/SegyIO',
      packages=['segyio'],
      package_dir={'': 'python'},
      package_data={'': ['License.md']},
      license='LGPL-3.0',
      ext_modules=[_segyio],
      platforms='any',
      requires=['numpy'],
      install_requires=['numpy'],
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Other Environment',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: C',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.5',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: Physics',
          'Topic :: Software Development :: Libraries',
          'Topic :: Utilities'
      ]
      )



from segyio import Enum


class BinField(Enum):
    JobID = 3201
    LineNumber = 3205
    ReelNumber = 3209
    Traces = 3213
    AuxTraces = 3215
    Interval = 3217
    IntervalOriginal = 3219
    Samples = 3221
    SamplesOriginal = 3223
    Format = 3225
    EnsembleFold = 3227
    SortingCode = 3229
    VerticalSum = 3231
    SweepFrequencyStart = 3233
    SweepFrequencyEnd = 3235
    SweepLength = 3237
    Sweep = 3239
    SweepChannel = 3241
    SweepTaperStart = 3243
    SweepTaperEnd = 3245
    Taper = 3247
    CorrelatedTraces = 3249
    BinaryGainRecovery = 3251
    AmplitudeRecovery = 3253
    MeasurementSystem = 3255
    ImpulseSignalPolarity = 3257
    VibratoryPolarity = 3259
    Unassigned1 = 3261
    SEGYRevision = 3501
    TraceFlag = 3503
    ExtendedHeaders = 3505
    Unassigned2 = 3507



"""
simple segy input/output

Welcome to segyio. For help, examples and reference, type `help(function)` in
your favourite python interpreter, or `pydoc function` in the unix console.

The segy library attempts to be easy to use efficently for prototyping and
interaction with possibly large segy files. File reading and writing is
streaming, with large file support out of the box and without hassle. For a
quick start on reading files, type `help(segyio.open)`.

An open segy file is interacted with in modes, found in the segy module. For a
reference with examples, please type `help(segyio.segy)`. For documentation on
individual modes, please refer to the individual modes with
`help(segyio.SegyFile.[mode])`, or look it up in the aggregated segyio.segy.
The available modes are:
    * text, for textual headers including extended headers
    * bin, for the binary header
    * header, for the trace headers
    * trace, for trace data
    * iline, for inline biased operations
    * xline, for crossline biased operations

The primary data type is the numpy array. All examples use `np` for the numpy
namespace. That means that any function that returns a trace, a set of samples
or even full lines, returns a numpy array. This enables quick and easy
mathematical operations on the data you care about.

Segyio is designed to blend into regular python code, so python concepts that
map to segy operations are written to behave similarly. That means that
sequences of data support list lookup, slicing (`f.trace[0:10:2]`), `for x in`
etc. Please refer to the individual mode's documentation for a more exhaustive
list with examples.

For all slicing operations that segyio provides the underlying buffer is reused,
so if you want to keep the data between iterations it is necessary to manually
copy the data. Please refer to the examples. (e.g. numpy.copy())
"""


class Enum(object):
    def __init__(self, enum_value):
        super(Enum, self).__init__()
        self._value = int(enum_value)

    def __int__(self):
        return int(self._value)

    def __str__(self):
        for k, v in self.__class__.__dict__.items():
            if isinstance(v, int) and self._value == v:
                return k
        return "Unknown Enum"

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(self._value)

    def __eq__(self, other):
        try:
            o = int(other)
        except ValueError:
            return super(Enum, self).__eq__(other)
        else:
            return self._value == o

    def __ne__(self, other):
        return not self == other

    @classmethod
    def enums(cls):
        result = []
        for k, v in cls.__dict__.items():
            if isinstance(v, int):
                result.append(cls(v))

        return sorted(result, key=int)


from .segysampleformat import SegySampleFormat
from .tracesortingformat import TraceSortingFormat
from .tracefield import TraceField
from .binfield import BinField
from .open import open
from .create import create
from .segy import SegyFile, spec
from .tools import dt, sample_indexes, create_text_header, native
from .tools import collect, cube

__version__    = '1.0.4'
__copyright__  = 'Copyright 2016, Statoil ASA'
__license__    = 'GNU Lesser General Public License version 3'
__status__     = 'Production'



import segyio
import numpy as np
import itertools as itr

def dt(segyfile, fallback_dt=4):
    """
    Find a *dt* value in the SegyFile. If none is found use the provided *fallback_dt* value.

    :type segyfile: segyio.SegyFile
    :type fallback_dt: float
    :rtype: float
    """
    return segyio._segyio.get_dt(segyfile.xfd, fallback_dt)


def sample_indexes(segyfile, t0=0.0, dt_override=None):
    """
    Creates a list of values representing the samples in a trace at depth or time.
    The list starts at *t0* and is incremented with am*dt* for the number of samples.
    If a *dt_override* is not provided it will try to find a *dt* in the file.

    :type segyfile: segyio.SegyFile
    :type t0: float
    :type dt_override: float or None
    :rtype: list[float]
    """
    if dt_override is None:
        dt_override = dt(segyfile)

    return [t0 + t * dt_override for t in range(len(segyfile.samples))]


def create_text_header(lines):
    """
    Will create a "correct" SEG-Y textual header.
    Every line will be prefixed with C## and there are 40 lines.
    The input must be a dictionary with the line number[1-40] as a key.
    The dictionary only requires entries for lines with content.
    The content for each line is a string with a maximum of 76 characters.

    :type lines: dict[int, str]
    :rtype: str
    """
    rows = []
    for line_no in range(1, 41):
        line = ""
        if line_no in lines:
            line = lines[line_no]
        row = "C{0:>2} {1:76}".format(line_no, line)
        rows.append(row)

    rows = ''.join(rows)
    return rows

def native(data,
           format = segyio.SegySampleFormat.IBM_FLOAT_4_BYTE,
           copy = True):
    """ Convert numpy array to native float

    :type data: numpy.ndarray
    :type format: int|segyio.SegySampleFormat
    :type copy: bool
    :rtype: numpy.ndarray

    Converts a numpy array from raw segy trace data to native floats. Works for numpy ndarrays.

    Examples:
        Convert mmap'd trace to native float:
        >>> d = np.memmap('file.sgy', offset = 3600, dtype = np.uintc)
        >>> samples = 1500
        >>> trace = segyio.tools.native(d[240:240+samples])
    """

    data = data.view( dtype = np.single )
    if copy:
        data = np.copy( data )

    format = int(segyio.SegySampleFormat(format))
    return segyio._segyio.native(data, format)

def collect(itr):
    """ Collect traces or lines into one ndarray

    Eagerly copy a series of traces, lines or depths into one numpy ndarray. If
    collecting traces or fast-direction over a post-stacked file, reshaping the
    resulting array is equivalent to calling `tools.cube`.

    Examples:

    collect-cube identity::
        >>> f = segyio.open('post-stack.sgy')
        >>> x = segyio.tools.collect(f.traces[:])
        >>> x = x.reshape((len(f.ilines), len(f.xlines), f.samples))
        >>> numpy.all(x == segyio.tools.cube(f))

    :type itr: iterable[numpy.ndarray]
    :rtype: numpy.ndarray
    """
    return np.stack([np.copy(x) for x in itr])

def cube(f):
    """ Read a full cube from a file

    Takes an open segy file (created with segyio.open) or a file name.

    If the file is a prestack file, the cube returned has the dimensions
    (fast,slow,offset,sample). If it is post-stack (i.e. only the one offset),
    the dimensions are normalised to (fast,slow,sample)

    :type f: SegyFile|str
    :rtype numpy.ndarray
    """

    if not isinstance(f, segyio.SegyFile):
        with segyio.open(f) as fl:
            return cube(fl)

    ilsort = f.sorting == segyio.TraceSortingFormat.INLINE_SORTING
    fast = f.ilines if ilsort else f.xlines
    slow = f.xlines if ilsort else f.ilines
    fast, slow, offs = len(fast), len(slow), len(f.offsets)
    smps = len(f.samples)
    dims = (fast, slow, smps) if offs == 1 else (fast, slow, offs, smps)
    return f.trace.raw[:].reshape(dims)
